@@ -11,7 +11,38 @@ def _parse_fraction(val: str) -> float:
         return float(a) / float(b)
     return float(val)
 
-def _parse_position(val: str) -> list[int]:
+def _parse_range(val: str) -> tuple[float, float]:
+    if ":" not in val:
+        return (0, _parse_fraction(val))
+    split = val.split(":")
+    if len(split) != 2:
+        raise ValueError("Must be in the form 'max' or 'min:max'")
+    try:
+        min = _parse_fraction(split[0])
+    except ValueError:
+        raise ValueError("Error parsing minimum")
+    try:
+        max = _parse_fraction(split[1])
+    except ValueError:
+        raise ValueError("Error parsing maximum")
+    return (min, max)
+
+def _parse_xy_range(val: str) -> tuple[tuple[float, float], tuple[float, float]]:
+    split = val.split(",")
+    if len(split) != 2:
+        raise ValueError("Must be in the form X_RANGE,Y_RANGE")
+    try:
+        x = _parse_range(split[0])
+    except ValueError:
+        raise ValueError("Error parsing x range")
+    try:
+        y = _parse_range(split[1])
+    except ValueError:
+        raise ValueError("Error parsing y range")
+    return np.array((x, y)).transpose()
+
+
+def _parse_position(val: str) -> tuple[float, float, float]:
     split = val.split(",")
     if len(split) != 3:
         raise ValueError("Must be in the form x,y,t")
@@ -27,7 +58,7 @@ def _parse_position(val: str) -> list[int]:
         t = _parse_fraction(split[2])
     except ValueError:
         raise ValueError("Error parsing t")
-    return [x,y,t]
+    return (x,y,t)
 
 def _movement_helper(data: synth_format.DataContainer, base_func, relative_func, pivot_func, relative: bool, pivot: list[int], *args, **kwargs) -> None:
     """pick the right function depending on relative or pivot being set"""
@@ -70,6 +101,7 @@ def get_parser():
     parser.add_argument("--invert-filter", action="store_true", help="Invert filter so everything *but* the filter is affected")
 
     preproc_group = parser.add_argument_group("pre-processing")
+    preproc_group.add_argument("-b", "--bpm", type=_parse_fraction, help="Change BPM without changing timing")
     preproc_group.add_argument("--delete-others", action="store_true", help="Delete everything that doesn't match the filter")
     preproc_group.add_argument("--connect-singles", type=_parse_fraction, metavar="MAX_INTERVAL", help="Replace strings of single notes with rails if they are . Use with'--rails-to-singles=1' if you want to keep the singles")
     preproc_group.add_argument("--merge-rails", type=_parse_fraction, nargs="?", action="append", metavar="MAX_INTERVAL", help="Merge sequential rails. By default only joins rails that start close to where another ends (in X, Y AND time), but with MAX_INTERVAL only the time interval counts")
@@ -93,7 +125,7 @@ def get_parser():
     movement_group.add_argument("--outset", type=_parse_fraction, metavar="DISTANCE", help="Move outwards")
 
     movement_group.add_argument("-c", "--stack-count", type=int, help="Instead of moving, create copies. Must have time offset set.")
-
+    movement_group.add_argument("--offset-random", type=_parse_xy_range, metavar="[MIN_X:]MAX_X,[MIN_Y:]MAX_Y", help="Offset by a random amount in the X axis")
 
     postproc_group = parser.add_argument_group("post-processing")
     postproc_group.add_argument("--split-rails", action="store_true", help="Split rails at single notes")
@@ -121,6 +153,10 @@ def main(options):
         abort(f"Could not decode clipboard, did you copy somethinge else?\n\t{err!r}")
 
     # preprocessing
+    if options.bpm:
+        bpm_scale = [1, 1, options.bpm / data.bpm]
+        data.apply_for_all(movement.scale, bpm_scale, types=filter_types)
+
     if options.delete_others:
         data = data.filtered(types=filter_types)
     if options.connect_singles:
@@ -141,6 +177,14 @@ def main(options):
         changed |= getattr(data, options.change_notes)
         setattr(data, options.change_notes, changed)
 
+    if options.change_walls:
+        new_type = synth_format.WALL_TYPES[options.change_walls][0]
+        def _change_wall_type(wall: "numpy array (1, 5)") -> "numpy array (1, 5)":
+            new_wall = wall.copy()
+            new_wall[..., 3] = new_type
+            return new_wall
+        data.apply_for_walls(_change_wall_type, types=filter_types)
+
     # rail patterns
     if options.interpolate:
         data.apply_for_notes(rails.interpolate_nodes_linear, options.interpolate, types=filter_types)
@@ -168,9 +212,18 @@ def main(options):
         stacking = data.filtered(types=filter_types)
         for i in range(options.stack_count):
             do_movement(options, stacking)
-            data.merge(stacking)
+            if options.offset_random is not None:
+                tmp = stacking.filtered()  # deep copy
+                random_offset = pattern_generation.random_xy(1, options.offset_random[0], options.offset_random[1])[0]
+                tmp.apply_for_all(movement.offset, [random_offset[0], random_offset[1], 0])
+                data.merge(tmp)
+            else:
+                data.merge(stacking)
     else:
         do_movement(options, data, filter_types=filter_types)
+        if options.offset_random is not None:
+            random_offset = pattern_generation.random_xy(1, options.offset_random[0], options.offset_random[1])[0]
+            data.apply_for_all(movement.offset, [random_offset[0], random_offset[1], 0])
 
     # postprocessing
     if options.split_rails:
